package org.jboss.cache.util.internals.replicationlisteners;

import org.jboss.cache.*;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.commands.ReplicableCommand;
import org.jboss.cache.commands.VisitableCommand;
import org.jboss.cache.commands.remote.*;
import org.jboss.cache.commands.tx.PrepareCommand;
import org.jboss.cache.factories.ComponentRegistry;
import org.jboss.cache.io.ByteBuffer;
import org.jboss.cache.marshall.AbstractMarshaller;
import org.jboss.cache.marshall.CommandAwareRpcDispatcher;
import org.jboss.cache.marshall.InactiveRegionAwareRpcDispatcher;
import org.jboss.cache.marshall.Marshaller;
import org.jboss.cache.marshall.RegionalizedMethodCall;
import org.jboss.cache.util.TestingUtil;
import org.jgroups.blocks.RpcDispatcher;
import org.jgroups.util.Buffer;
import org.jgroups.Address;

import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.*;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

/**
 * Utility class that notifies when certain commands were asynchronously replicated on secondary cache.
 * Especially useful for avaoiding Thread.sleep() statements.
 * <p/>
 * Usage:
 * <pre>
 *   no tx:
 *       Cache c1, c2; //these being two async caches
 *       ReplicationListener listener2 = ReplicationListener.getReplicationListener(c2);
 *       listener2.expect(PutKeyValueCommand.class);
 *       c1.put(fqn, key, value);
 *       listener2.waitForReplicationToOccur(1000); // -this will block here untill c2 recieves the PutKeyValueCommand command
 *   with tx: (difference is that expectWithTx is used insted of expect
 *       Cache c1, c2; //these being two async caches
 *       ReplicationListener listener2 = ReplicationListener.getReplicationListener(c2);
 *       listener2.expectWithTx(PutKeyValueCommand.class);
 *       txManager.begin();
 *       c1.put(fqn, key, value);
 *       txManager.commit();
 *       listener2.waitForReplicationToOccur(1000); // -this will block here untill c2 recieves the PutKeyValueCommand command
 * </pre>
 *
 *
 * Lifecycle - after being used (i.e. waitForReplicationToOccur returns sucessfully) the object returns to the
 * non-initialized state and *can* be reused through expect-wait cycle.
 * <b>Note</b>:  this class might be used aswell for sync caches, e.g. a test could have subclasses which use sync and
 * async replication
 *
 * @author Mircea.Markus@jboss.com
 * @since 2.2
 */
abstract public class ReplicationListener
{

   public static final long DEFAULT_TIMEOUT = 10000;
   private CountDownLatch latch = new CountDownLatch(1);
   protected List<Class<? extends ReplicableCommand>> expectedCommands;
   protected Configuration config;
   protected final Address localAddress;

   /**
    * Builds a listener that will observe the given cache for recieving replication commands.
    */
   protected ReplicationListener(Cache cache)
   {
      ComponentRegistry componentRegistry = TestingUtil.extractComponentRegistry(cache);
      RPCManager rpcManager = componentRegistry.getComponent(RPCManager.class);
      CommandAwareRpcDispatcher realDispatcher = (CommandAwareRpcDispatcher) TestingUtil.extractField(rpcManager, "rpcDispatcher");
      RpcDispatcher.Marshaller2 realMarshaller = (RpcDispatcher.Marshaller2) realDispatcher.getMarshaller();
      RpcDispatcher.Marshaller2 delegate = null;
      if ((realMarshaller instanceof RegionMarshallerDelegate) || (realMarshaller instanceof MarshallerDelegate))
      {
         throw new RuntimeException("Illegal state");
      }
      if (realDispatcher instanceof InactiveRegionAwareRpcDispatcher)
         delegate = new RegionMarshallerDelegate((Marshaller) realMarshaller);
      else
         delegate = new MarshallerDelegate(realMarshaller);
      realDispatcher.setMarshaller(delegate);
      realDispatcher.setRequestMarshaller(delegate);
      realDispatcher.setResponseMarshaller(delegate);
      this.config = cache.getConfiguration();
      this.localAddress = cache.getLocalAddress();
   }

   protected ReplicationListener()
   {
      localAddress = null;
   }

   abstract public void expect(Class<? extends ReplicableCommand>... expectedCommands);

   /**
    * Based on cache's configuration, will know for what specific commands to expect to be replicated.
    * E.g. async replication with a tx, would expect only a PrepareCommand (async is 1PC). sync repl with tx would expect
    * a prepare and a commit (sync is 2pc).
    */
   abstract public void expectWithTx(Class<? extends ReplicableCommand>... writeCommands);

   /**
    * Factory method, to be used in order to obtain a replication listener based on a cache config.
    */
   public static ReplicationListener getReplicationListener(Cache cache)
   {
      if (cache.getConfiguration().getCacheMode().isInvalidation())
      {
         return new InvalidationReplicationListener(cache);
      }
      if (cache.getConfiguration().getNodeLockingScheme().equals(Configuration.NodeLockingScheme.OPTIMISTIC))
      {
         return new OptimisticReplicationListener(cache);
      }
      else if (cache.getConfiguration().getNodeLockingScheme().equals(Configuration.NodeLockingScheme.PESSIMISTIC))
      {
         return new PessReplicationListener(cache);
      } else
      {
         return new MvccReplicationListener(cache);
      }
   }

   private class MarshallerDelegate implements RpcDispatcher.Marshaller2
   {
      RpcDispatcher.Marshaller2 marshaller;

      private MarshallerDelegate(RpcDispatcher.Marshaller2 marshaller)
      {
         this.marshaller = marshaller;
      }

      public byte[] objectToByteBuffer(Object obj) throws Exception
      {
         return marshaller.objectToByteBuffer(obj);
      }

      public Object objectFromByteBuffer(byte bytes[]) throws Exception
      {
         Object result = marshaller.objectFromByteBuffer(bytes);
         return process(result);
      }

      public Object objectFromByteBuffer(byte[] bytes, int i, int i1) throws Exception
      {
         Object result = marshaller.objectFromByteBuffer(bytes, i, i1);
         return process(result);
      }

      private Object process(Object result)
      {
         System.out.println(localAddress + " >>>> " + result );
         boolean isReplicate = result instanceof ReplicateCommand;
         boolean isRemote = result != null && isRemoteCommand(result.getClass());
         if ((isReplicate || isRemote) && (expectedCommands != null))
         {
            ReplicableCommand replicateCommand = (ReplicableCommand) result;
            result = new ReplicableCommandDelegate(replicateCommand);
         }
         return result;
      }

      public Buffer objectToBuffer(Object o) throws Exception
      {
         return marshaller.objectToBuffer(o);
      }
   }

   /**
    * We want the notification to be performed only *after* the remote command is executed.
    */
   private class ReplicableCommandDelegate implements ReplicableCommand
   {
      ReplicableCommand realOne;

      private ReplicableCommandDelegate(ReplicableCommand realOne)
      {
         if (realOne instanceof VisitableCommand)
         {
            throw new IllegalArgumentException("Visitable commands not allowed!!! ;recieved="
                  + realOne.getClass().getName());
         }
         this.realOne = realOne;
      }

      public Object perform(InvocationContext ctx) throws Throwable
      {
         try
         {
            return realOne.perform(ctx);
         }
         finally
         {
            if (realOne instanceof ReplicateCommand)
            {
               postReplicateExecution((ReplicateCommand)realOne);
            }
            else
            {
               postNonVisitableExecution(realOne);
            }
            if (expectedCommands.isEmpty())
            {
               latch.countDown();
            }
         }
      }

      public int getCommandId()
      {
         return realOne.getCommandId();
      }

      public Object[] getParameters()
      {
         return realOne.getParameters();
      }

      public void setParameters(int commandId, Object[] parameters)
      {
         realOne.setParameters(commandId, parameters);
      }
   }

   protected void postNonVisitableExecution(ReplicableCommand realOne)
   {
      expectedCommands.remove(realOne.getClass());
   }
   
   protected void postReplicateExecution(ReplicateCommand realOne)
   {
      System.out.println("Processed command: " + realOne + "; expecting " + expectedCommands.size() + " commands");
      Iterator<Class<? extends ReplicableCommand>> it = expectedCommands.iterator();
      while (it.hasNext())
      {
         Class<? extends ReplicableCommand> replicableCommandClass = it.next();
         if (realOne.containsCommandType(replicableCommandClass))
         {
            it.remove();
         } else if (realOne.getSingleModification() instanceof PrepareCommand) //explicit transaction
         {
            PrepareCommand prepareCommand = (PrepareCommand) realOne.getSingleModification();
            if (prepareCommand.containsModificationType(replicableCommandClass))
            {
               it.remove();
            }
         }
      }
      System.out.println("After process expecing : " + expectedCommands.size() + " commands");
   }

   /**
    * Needed for region based marshalling.
    */
   private class RegionMarshallerDelegate extends AbstractMarshaller
   {
      private Marshaller realOne;

      private RegionMarshallerDelegate(Marshaller realOne)
      {
         this.realOne = realOne;
      }

      public void objectToObjectStream(Object obj, ObjectOutputStream out) throws Exception
      {
         realOne.objectToObjectStream(obj, out);
      }

      public Object objectFromObjectStream(ObjectInputStream in) throws Exception
      {
         return realOne.objectFromObjectStream(in);
      }

      public Object objectFromStream(InputStream is) throws Exception
      {
         return realOne.objectFromStream(is);
      }

      public void objectToObjectStream(Object obj, ObjectOutputStream out, Fqn region) throws Exception
      {
         realOne.objectToObjectStream(obj, out, region);
      }

      public RegionalizedMethodCall regionalizedMethodCallFromByteBuffer(byte[] buffer) throws Exception
      {
         RegionalizedMethodCall result = realOne.regionalizedMethodCallFromByteBuffer(buffer);
         if (result.command instanceof ReplicateCommand && expectedCommands != null)
         {
            ReplicateCommand replicateCommand = (ReplicateCommand) result.command;
            result.command = new ReplicableCommandDelegate(replicateCommand);
         }
         return result;
      }

      public RegionalizedMethodCall regionalizedMethodCallFromObjectStream(ObjectInputStream in) throws Exception
      {
         return realOne.regionalizedMethodCallFromObjectStream(in);
      }

      public ByteBuffer objectToBuffer(Object o) throws Exception
      {
         return realOne.objectToBuffer(o);
      }

      public Object objectFromByteBuffer(byte[] bytes, int i, int i1) throws Exception
      {
         return realOne.objectFromByteBuffer(bytes, i, i1);
      }
   }

   /**
    * Blocks for the elements specified through {@link #expect(Class[])} invocations to be replicated in this cache.
    * if replication does not occur in the give timeout then an exception is being thrown.
    */
   public void waitForReplicationToOccur(long timeoutMillis)
   {
      System.out.println("enter... ReplicationListener.waitForReplicationToOccur");
      waitForReplicationToOccur(timeoutMillis, TimeUnit.MILLISECONDS);
      System.out.println("exit... ReplicationListener.waitForReplicationToOccur");
   }

   /**
    * same as {@link #waitForReplicationToOccur(long)}, just that it uses the {@link #DEFAULT_TIMEOUT} for timeout.
    */
   public void waitForReplicationToOccur()
   {
      waitForReplicationToOccur(DEFAULT_TIMEOUT);
   }

   /**
    * Similar to {@link #waitForReplicationToOccur(long)} except that this method provides more flexibility in time units.
    *
    * @param timeout  the maximum time to wait
    * @param timeUnit the time unit of the <tt>timeout</tt> argument.
    */
   public void waitForReplicationToOccur(long timeout, TimeUnit timeUnit)
   {
      assert expectedCommands != null : "there are no replication expectations; please use AsyncReplicationListener.expectWithTx(...) before calling this method";
      try
      {
         if (!expectedCommands.isEmpty() && !latch.await(timeout, timeUnit))
         {
            assert false : "waiting for more than " + timeout + " " + timeUnit + " and following commands did not replicate: " + expectedCommands;
         }
      }
      catch (InterruptedException e)
      {
         throw new IllegalStateException("unexpected", e);
      }
      finally
      {
         expectedCommands = null;
         latch = new CountDownLatch(1);
      }
   }

   /**
    * {@link #waitForReplicationToOccur(long)} will block untill all the commands specified here are being replicated
    * to this cache. The method can be called several times with various arguments.
    */
   protected void internalExpect(Class<? extends ReplicableCommand>... expectedCommands)
   {
      if (this.expectedCommands == null)
      {
         this.expectedCommands = new ArrayList<Class<? extends ReplicableCommand>>();
      }
      this.expectedCommands.addAll(Arrays.asList(expectedCommands));
   }

   protected boolean isRemoteCommand(Class clazz)
   {
      return clazz.equals(AnnounceBuddyPoolNameCommand.class) || clazz.equals(AssignToBuddyGroupCommand.class) ||
            clazz.equals(ClusteredGetCommand.class) || clazz.equals(DataGravitationCleanupCommand.class) ||
            clazz.equals(RemoveFromBuddyGroupCommand.class) || clazz.equals(ReplicateCommand.class);
   }
}
